//TRS-80 model 1 cassette converter
// Copyright (c)2000 Knut Roll-Lund
// updated to wav2cas_l2 to distinguish it from level 1 (wav2cas_l1)
//
// Converts a general wav file into cas format supported by emulators
// The wave file can be any sample rate, 16bit, 8 bit and even stereo
// it will assume data to be readable by the bios programs, 500 baud
// (high speed or 250 baud won't work due to filtering
// aimed at 500 baud, and lack of information on my part)
// The data is faithfully registered in from the first clock
// pulses detected and aligned with the A5 synch in a similar fasion
// as the bios, from then on bytes will be converted until "silence"
// is detected and a new synchronization round is done if more data comes
//
// As I expect there might be an interest in converting this program
// for use with other baudrates etc. This source is awailable
// for download and can be modified and used so long as:
// acknowledgement of my effort is made, my copyright notice
// is not removed, and not part of a commersial product.
//
// If you want to use the code in a commersial product contact me
// mailto:kr-lund@online.no www:http://home.online.no/~kr-lund
// also I will probably continue to be reading
// news:comp.sys.tandy

#include <windows.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "resource.h"

typedef unsigned char byte;
typedef unsigned short word;
typedef unsigned long dword;

#define TRUE 1
#define FALSE 0

struct {
   char name[4];           //'RIFF'
   dword size;             //filesize-8
} riff;

struct {
   word tag;               //0x0001
   word channels;          //whatever
   dword samples_per_sec;  //whatever
   dword bytes_per_sec;    //samples_per_sec*bytes_per_sample*channels
   word block;             //bits_per_sample*channels/8
   word bits_per_sample;   //whatever (per channel)
   word size;              //extention data (new format WAVEFORMATEX)
} fmt;

////total header for wave output
//#pragma pack(1)
//struct {
//   char name1[4];          //riff
//   dword size1;
//   char name2[4];          //wave
//   char name3[4];          //fmt
//   dword size2;
//   word tag;
//   word channels;
//   dword samples_per_sec;
//   dword bytes_per_sec;
//   word block;
//   word bits_per_sample;
//   word exsize;
//   char name4[4];          //data
//   dword size3;
//} wave = {
//   {'R','I','F','F'},  //MMheader
//   38,      //size (samples*2+38)
//   {'W','A','V','E'},  //wave header
//   {'f','m','t',' '},  //format header
//   18,      //size of format
//   1,       //tab WAVE_FORMAT_PCM
//   1,       //channels (mono)
//   0,       //samples per second (get from original)
//   0,       //bytes per second (calc from original)
//   2,       //block -> word format
//   16,      //bits per sample -> word
//   0,       //extention data size = 0
//   {'d','a','t','a'},  //data header
//   0        //size of data (samples*2)
//};
//#pragma pack()

//int main(int argc,char *argv[])
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE dummy, LPSTR lpCmdLine, int nCmdShow )
{
   char drive[_MAX_DRIVE], dir[_MAX_DIR], file[_MAX_FNAME], ext[_MAX_EXT];
   char path1[_MAX_PATH+50], path2[_MAX_PATH];
   dword size, n;
   byte *buff1;
   byte *p;
   short *buff2;
   short *q;
   long i;
   short j,min,max,lim;
   byte d_beg,d_end,n_beg;
   byte c,d,f;
   byte synch;
   FILE *ifil, *ofil/*, *xfil*/;

   char *argv[3];
   int argc;
   char *pek;

   argc = 1;
   argv[1] =
   argv[2] = NULL;

   //MessageBox(NULL,lpCmdLine,"WAV2CAS_L2 debug",MB_OK | MB_ICONERROR);
   //scan for arguments
   for (pek=lpCmdLine; *pek!=0;)
   {  //skip extra whitespace
      while ((*pek==' ')||(*pek=='\t'))
         pek++;
      if (*pek==0)      //end of string
         break;
      if (*pek!='\"')   //quote
      {
         argv[argc] = pek;
         argc++;
         //look for space, tab or end of string
         while ((*pek!=' ')&&(*pek!='\t')&&(*pek!=0))
            pek++;
         if (*pek!=0)
            *pek++ = 0;
      }
      else
      {
         pek++;   //bump past quote
         argv[argc] = pek;
         argc++;
         //look for quote or end of string
         while ((*pek!='\"')&&(*pek!=0))
            pek++;
         if (*pek!=0)
            *pek++ = 0;
      }
      if (argc==3)   //ignore more arguments
         break;
   }
   if (argc==1)
   {
      MessageBox(
         NULL,
         "WAV2CAS_L2 TRS-80 cassette coverter 1.0\n"
         "2000 Knut Roll-Lund\n\n"
         "Use:\n"
         "WAV2CAS_L2 InFile [OutFile]\n\n"
         "InFile filename and path of the prerecorded WAV file\n"
         "   default extention .WAV\n"
         "OutFile optional filename of output file\n"
         "   default same name with extention .CAS\n\n"
		 "Note: If you create a shortcut on the desktop you\n"
		 "can use drag&drop, dropping wav files on the icon.",
         "About WAV2CAS_L2",
         MB_OK );
      return 0;
   }

   //establish name of InFile
   _splitpath((char*)argv[1],(char*)drive,(char*)dir,(char*)file,(char*)ext);
   if (*ext==0)
      strcpy((char*)ext,".WAV");
   _makepath( (char*)path1,(char*)drive,(char*)dir,(char*)file,(char*)ext );

   //open InFile
   if (( ifil= fopen( (char*)path1, "r+b" ))==NULL)
   {
      //printf("Cannot open %s!",path1);
      strcat(path1,"\nUnable to open file!");
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 2;
   }

   //read riff header
   if (fread( &riff, 1, 8, ifil )!=8)
   {
      //printf("%s is not a proper WAV file!",path1);
      fclose( ifil );
      strcat(path1,"\nis not a proper WAV file!");
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 2;
   }

   //check riff header
   if ((memcmp(riff.name,"RIFF",4)!=0)||(riff.size<=4+8+16+8))
   {
      //printf("%s is not a proper WAV file!",path1);
      fclose( ifil );
      strcat(path1,"\nis not a proper WAV file!");
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 2;
   }

   //grab the size of the riff file
   size = riff.size;

   //find the wave :-)
   while (1)
   {
      fread( &riff, 1, 4, ifil );
      if (memcmp( riff.name, "WAVE", 4 )==0)
         break;   //out of while
      //not wave; read size too
      fread( &riff.size, 1, 4, ifil );
      //try skipping this block (could only be a LIST block, I think)
      if ((fseek(ifil, riff.size, SEEK_CUR)!=0)||(size<=riff.size+4+8+16+8))
      {
         //printf("%s is not a proper WAV file!",path1);
         fclose( ifil );
         strcat(path1,"\nis not a proper WAV file!");
         MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
         return 2;
      }
      size -= 8+riff.size;
   }
   //read fmt header
   fread( &riff, 1, 8, ifil );

   //check for correctness
   if (memcmp( riff.name, "fmt ", 4 )!=0)
   {
      //printf("%s is not a proper WAV file!",path1);
      fclose( ifil );
      strcat(path1,"\nis not a proper WAV file!");
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 2;
   }
   if (riff.size>18)
   {
      //printf("Extra data in the fmt header\n%s must be an uncompressed and otherwise straight WAV file!",path1);
      fclose( ifil );
      strcat(path1,"\nextra data in the fmt header!");
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 2;
   }
   //read the fmt data into the buffer supporting both old and present format
   fread( &fmt, 1, riff.size, ifil );

   //adjust size for the whole wave construct without the data
   size -= 4+8+riff.size+8;

   //get the data chunk header
   fread( &riff, 1, 8, ifil );

   //check data
   if (memcmp( riff.name, "data", 4 )!=0)
   {
      //printf("%s is not a proper WAV file!",path1);
      fclose( ifil );
      strcat(path1,"\nis not a proper WAV file!");
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 2;
   }

   //we now have a valid InFile positioned at start of data

   //establish name of OutFile
   if (argc>2)
   {
      _splitpath((char*)argv[2],(char*)drive,(char*)dir,(char*)file,(char*)ext);
      if (*ext==0)
         strcpy((char*)ext,".CAS");
   }
   else
      strcpy((char*)ext,".CAS");
   _makepath( (char*)path2,(char*)drive,(char*)dir,(char*)file,(char*)ext );

   //open OutFile
   if (( ofil= fopen( (char*)path2, "w+b" ))==NULL)
   {
      //printf("Cannot open/create %s!",path2);
      fclose( ifil );
      sprintf(path1,"Cannot open/create %s!",path2);
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 3;
   }

   //calculate number of samples to process
   size = riff.size/fmt.block;

   //allocate input buffer one minute (worst case 48000*4*60+4=11520004) and filter buffer (48000*2*60=570000)
   if (((buff1=malloc(riff.size))==NULL)||((buff2=malloc((size-1)*2))==NULL))
   {
      //printf("Unable to allocate buffers (0x%08x)",riff.size+(size-1)*2);
      fclose( ifil );
      fclose( ofil );
      free( buff1 );
      free( buff2 );
      sprintf(path1,"Unable to allocate buffers (0x%08x)",riff.size+(size-1)*2);
      MessageBox(NULL,path1,"WAV2CAS_L2 Error",MB_OK | MB_ICONERROR);
      return 4;
   }

   //calculate times for decoding 500baud
   d_beg = (byte)(fmt.samples_per_sec/1050); //placement of datapulse from clock pulse .95ms in samples (44100=>42)
   d_end = (byte)(fmt.samples_per_sec/816);  //placement of datapulse from clock pulse 1.22ms in samples (44100=>54)
   n_beg = (byte)(fmt.samples_per_sec/512);  //placement of the next clock pulse 1.95ms in samples (44100=>86)

   //prime min and max
   min = 32767;
   max = -32768;

   //read the whole sh'bang
   fread( buff1, 1, riff.size, ifil );

   //form here handle size-1 samples (working with the difference makes us "loose" 1 sample)
   size--;

   //find difference and combining stereo
   for (p=buff1, q=buff2, n=size; n!=0; n--)
   {
      if (fmt.channels>1)
         if (fmt.bits_per_sample>8)
         {  //16bit stereo
            i = ((long)*((short *)p)++)+((long)*((short *)p)++);     //add left and right of this sample
            i -= ((long)*((short *)p))+((long)*((short *)(p+2)));    //find diff to next sample
            j = (short)(i/2);   //average channels (so they fit in a short)
         }
         else
         {  //8bit stereo
            j = ((short)*p++)+((short)*p++);    //add left and right of this sample
            j -= ((short)*p)+((short)*(p+1));    //find diff to next sample
         }
      else
         if (fmt.bits_per_sample>8)
         {  //16bit mono
            j = (*((short *)p)++);
            j -= (*((short *)p));  //find diff to next sample
         }
         else
         {  //8bit mono
            j = (short)(*p++);     //get current sample
            j -= (short)(*p);      //subtract next sample
         }
      //move min/max if outside
      if (j < min)
         min = j;
      if (j > max)
         max = j;
      //store in the second buffer
      *q++ = j;
   }

   //discard the first buffer
   free( buff1 );

//   //write debug file 1 (diff and combine)
//   wave.size1 += size*2;
//   wave.size3 = size*2;
//   wave.samples_per_sec = fmt.samples_per_sec;
//   wave.bytes_per_sec = wave.samples_per_sec*2;
//   if (( xfil= fopen( "debug_1.wav", "w+b" ))!=NULL)
//   {
//      fwrite(&wave,1,sizeof(wave),xfil);
//      fwrite(buff2,1,wave.size3,xfil);
//      fclose(xfil);
//   }

   //setup trigger limit at 1 fourth of span
   lim = min/4;
//   printf("min:%d max:%d lim:%d\n",min,max,lim);

//   //write debug file 2 (diff, combine and evaluate)
//   if (( xfil= fopen( "debug_2.wav", "w+b" ))!=NULL)
//   {
//      max = 32767;
//      min = 0;
//      fwrite(&wave,1,sizeof(wave),xfil);
//      for(q=buff2, n=size; n!=0; n--)
//         if (*q++>=lim)
//            fwrite(&max,1,2,xfil);
//         else
//            fwrite(&min,1,2,xfil);
//      fclose(xfil);
//   }

   //analyze the data
   q = buff2;
   c = d = 0;
   synch = TRUE;  //look for synch byte 0xA5

   while (size)
   {
      //this should find the start of the clock pulse
      while ((*q>lim)&&(size!=0))
      {
         q++;
         size--;
      }
      if (size==0)
         break;
      if (size<=d_beg)
         break;            //data pulse in next batch
      if (size<d_end)      //end in the middle of the data pulse area
         d_end = (byte)size;
      d <<=1;
      for (f=d_beg; f<=d_end; f++)
         if (*(q+f)<=lim)
         {
            d |= 1;
            break;   //out of for
         }
      if (synch)
         if (d==0xA5)
         {
            synch = FALSE;
            c = 7;
         }
      if (++c==8)
      {
         if (synch)
         {
            fputc(0,ofil);
            c = 0;
         }
         else
         {
            fputc(d,ofil);
            c = d = 0;
         }
      }
      //find the next clock pulse
      if (size<n_beg)
         break;
      q += n_beg;
      size -= n_beg;
   }

   fclose( ofil );
   free( buff2 );
   sprintf(path1,"Conversion complete\nmin:%d max:%d lim:%d",-max,-min,-lim);
   MessageBox(NULL,path1,"WAV2CAS_L2 finished",MB_OK);
   return 0;
}